package buf

import (
	"bytes"
	"sync"
	"io"
)

type AtomicBuffers interface {
	Bytes() []byte
	Cap() int
	Grow(int)
	Len() int
	Next(int) []byte
	ReadByte() (byte, error)
	ReadBytes(byte) ([]byte, error)
	ReadFrom(io.Reader) (int64, error)
	ReadRune() (rune, int, error)
	ReadString(byte) (string, error)
	Reset()
	Truncate(int)
	UnreadByte() error
	UnreadRune() error
	WriteByte(byte) error
	WriteRune(rune) (int, error)
	WriteString(string) (int, error)
	WriteTo(io.Writer) (int64, error)
	Read([]byte) (int, error)
	Write([]byte) (int, error)
	String() string
}

type atomicBuffer struct {
	b bytes.Buffer
	m sync.RWMutex
}

func (b *atomicBuffer) Bytes() []byte {
	b.m.RLock()
	defer b.m.RUnlock()
	return b.b.Bytes()
}
func (b *atomicBuffer) Cap() int {
	b.m.RLock()
	defer b.m.RUnlock()
	return b.b.Cap()
}
func (b *atomicBuffer) Grow(n int) {
	b.m.Lock()
	defer b.m.Unlock()
	b.b.Grow(n)
}
func (b *atomicBuffer) Len() int {
	b.m.RLock()
	defer b.m.RUnlock()
	return b.b.Len()
}
func (b *atomicBuffer) Next(n int) []byte {
	b.m.Lock()
	defer b.m.Unlock()
	return b.b.Next(n)
}
func (b *atomicBuffer) ReadByte() (c byte, err error) {
	b.m.Lock()
	defer b.m.Unlock()
	return b.b.ReadByte()
}
func (b *atomicBuffer) ReadBytes(delim byte) (line []byte, err error) {
	b.m.Lock()
	defer b.m.Unlock()
	return b.b.ReadBytes(delim)
}
func (b *atomicBuffer) ReadFrom(r io.Reader) (n int64, err error) {
	b.m.Lock()
	defer b.m.Unlock()
	return b.b.ReadFrom(r)
}
func (b *atomicBuffer) ReadRune() (r rune, size int, err error) {
	b.m.Lock()
	defer b.m.Unlock()
	return b.b.ReadRune()
}
func (b *atomicBuffer) ReadString(delim byte) (line string, err error) {
	b.m.Lock()
	defer b.m.Unlock()
	return b.b.ReadString(delim)
}
func (b *atomicBuffer) Reset() {
	b.m.Lock()
	defer b.m.Unlock()
	b.b.Reset()
}
func (b *atomicBuffer) Truncate(n int) {
	b.m.Lock()
	defer b.m.Unlock()
	b.b.Truncate(n)
}
func (b *atomicBuffer) UnreadByte() error {
	b.m.Lock()
	defer b.m.Unlock()
	return b.b.UnreadByte()
}
func (b *atomicBuffer) UnreadRune() error {
	b.m.Lock()
	defer b.m.Unlock()
	return b.b.UnreadRune()
}
func (b *atomicBuffer) WriteByte(c byte) error {
	b.m.Lock()
	defer b.m.Unlock()
	return b.b.WriteByte(c)
}
func (b *atomicBuffer) WriteRune(r rune) (n int, err error) {
	b.m.Lock()
	defer b.m.Unlock()
	return b.b.WriteRune(r)
}
func (b *atomicBuffer) WriteString(s string) (n int, err error) {
	b.m.Lock()
	defer b.m.Unlock()
	return b.b.WriteString(s)
}
func (b *atomicBuffer) WriteTo(w io.Writer) (n int64, err error) {
	b.m.Lock()
	defer b.m.Unlock()
	return b.b.WriteTo(w)
}

func (b *atomicBuffer) Read(p []byte) (n int, err error) {
	b.m.Lock()
	defer b.m.Unlock()
	return b.b.Read(p)
}
func (b *atomicBuffer) Write(p []byte) (n int, err error) {
	b.m.Lock()
	defer b.m.Unlock()
	return b.b.Write(p)
}
func (b *atomicBuffer) String() string {
	b.m.RLock()
	defer b.m.RUnlock()
	return b.b.String()
}
